#ifndef LIBSQLITE_H
#define LIBSQLITE_H
/*
 * Do not introduce any kerio dependencies because the code is used in kerio asterisk modules
 */
#include <linux/limits.h>
#include <sys/file.h>
#include <fcntl.h>
#include <sqlite3.h>
#include <set>
#include <boost/foreach.hpp>

#ifdef AST_MODULE
#define ELOG(...) ast_log(AST_LOG_ERROR, __VA_ARGS__)
#define DLOG(...) ast_debug(1, __VA_ARGS__)
#else
#include "libKtsLog.h"
#define ELOG(...) eLog(__VA_ARGS__)
#define DLOG(...) ktsLog(DLOG_sqlwrite, __VA_ARGS__)
#endif

class SqliteLock {
	std::string lockFile;
	int fd;
	bool lockCalled;
public:

	SqliteLock(std::string lockFile) : lockFile(lockFile), fd(-1), lockCalled(false)
	{
	};

	~SqliteLock()
	{
		assert(lockCalled); /* lock method was not called, design error */
		::close(fd);
	};

	bool lock()
	{
		lockCalled = true;

		/* exclusive open, (blocking) */
		fd = ::open(lockFile.c_str(), O_RDWR | O_CREAT | O_APPEND, 0666);
		if (fd < 0) {
			ELOG("SqliteLock: Failed to create or open lock file %s.\n", lockFile.c_str());
			return false;
		}

		if (flock(fd, LOCK_EX) < 0) {
			ELOG("SqliteLock: Failed to lock file %s.\n", lockFile.c_str());
			::close(fd);
			fd = -1;
			return false;
		}

		return true;
	};

	/**
	 * disable the assert in destructor
	 */
	void ignoreLock()
	{
		lockCalled = true;
	}
};

class SqliteResult {
public:
	sqlite3_stmt *statement;

	SqliteResult() : statement(NULL)
	{
	}

	~SqliteResult()
	{
		reset();
	}

	void reset()
	{
		if (this->statement) {
			sqlite3_finalize(this->statement);
			this->statement = NULL;
		}
	}

};

class SqliteWrapper {
	const int sqliteRetries;
	const int sqliteUsleep;
	sqlite3 *sqlite;
	std::string database;
public:

	SqliteWrapper() : sqliteRetries(100), sqliteUsleep(100000), sqlite(NULL), database("")
	{
		srand(time(NULL));
	}

	~SqliteWrapper()
	{
		close();
	}

	void sqliteSleep()
	{
		usleep(sqliteUsleep / 2 + (random() % sqliteUsleep));
	}

	/**
	 * Open a sqlite database
	 *
	 * @param database
	 * @param eraseOnError
	 * @param synchronous - true(sync), false(no-sync)
	 * @param check - run a quick check
	 * @return [bool] true on success
	 */
	bool open(const std::string database, bool autoCreate = false, bool eraseOnError = false, bool synchronous = true, bool check = true)
	{
		if (this->sqlite) {
			return true; /* already opened */
		}

		int res = SQLITE_ERROR;

		for (int repeat = 0; repeat < this->sqliteRetries; repeat++) {
			/* open the database */
			res = sqlite3_open_v2(database.c_str(), &this->sqlite, SQLITE_OPEN_READWRITE | (autoCreate ? SQLITE_OPEN_CREATE : 0), NULL);
			if (res == SQLITE_OK) {
				break;
			}

			this->sqliteSleep();
		}
		this->database = database;

		if (check) {
			if (res == SQLITE_OK) {
				if (!this->exec("PRAGMA quick_check")) {
					ELOG("libSqlite(%s): Database check failed %d: %s.\n", database.c_str(), sqlite3_errcode(this->sqlite), sqlite3_errmsg(this->sqlite));
					this->close();
				}
			}
			else {
				ELOG("libSqlite(%s): Failed to open the database.\n", database.c_str());
				this->close();
			}
		}

		if (this->isOpen()) {
			/* database is open and passed the check */
			if (!synchronous) {
				this->exec("PRAGMA synchronous=OFF");
			}

			return true; /* success */
		}

		/* database is broken */
		this->close();
		this->database = "";

		if (eraseOnError) {
			/* erase and try again */
			DLOG("libSqlite(%s): Erasing the database.\n", database.c_str());
			this->dropDatabase(database);
			return this->open(database, autoCreate, false, synchronous);
		}

		return false; /* error */
	}

	/**
	 * Close a sqlite database
	 *
	 * The sqlite3_close may return SQLITE_BUSY. There is no error handling as
	 * there is really nothing to do when the database can't be closed.
	 */
	void close()
	{
		if (this->isOpen()) {
			int res = SQLITE_ERROR;
			for (int repeat = 0; repeat < this->sqliteRetries; repeat++) {
				res = sqlite3_close(this->sqlite);
				if (res != SQLITE_BUSY && res != SQLITE_LOCKED) {
					break; /* SQLITE_OK or an error */
				}

				DLOG("Sqlite::close - Failed to close database %s: %d, attempt %d.\n", this->database.c_str(), res, repeat);

				/* retry */
				this->sqliteSleep();
			}

			if (res != SQLITE_OK) {
				ELOG("Sqlite::close - Failed to close database %s: %d.\n", this->database.c_str(), res);
			}
			this->sqlite = NULL;
		}
	}

	inline bool isOpen()
	{
		return this->sqlite ? true : false;
	}

	/**
	 * Execute a sql command
	 *
	 * @param sqlFormat
	 * @param ...
	 * @return [bool] true on success
	 */
	bool exec(const char *sqlFormat, ...)
	{
		if (!this->isOpen()) {
			ELOG("libSqlite(%s): Database is not opened.\n", this->database.c_str());
			return false; /* failure */
		}

		/* prepare the sql query */
		char *sqlQuery;
		va_list sqlArgs;
		va_start(sqlArgs, sqlFormat);
		sqlQuery = sqlite3_vmprintf(sqlFormat, sqlArgs); /* must be freed */
		va_end(sqlArgs);

		int res = SQLITE_ERROR;
		for (int repeat = 0; repeat < this->sqliteRetries; repeat++) {
			res = sqlite3_exec(this->sqlite, sqlQuery, 0, 0, 0);
			if (res != SQLITE_BUSY && res != SQLITE_LOCKED) {
				break; /* SQLITE_OK or an error */
			}

			/* retry */
			this->sqliteSleep();
		}

		if (res != SQLITE_OK) {
			ELOG("libSqlite(%s): Failed exec \"%s\" error %d: %s.\n", this->database.c_str(), sqlQuery, sqlite3_errcode(this->sqlite), sqlite3_errmsg(this->sqlite));
			sqlite3_free(sqlQuery);
			return false; /* failure */
		}

		sqlite3_free(sqlQuery);

		return true; /* success */
	};

	/**
	 * Execute a sql query and return the result
	 * Use queryFetch to access the records
	 *
	 * @param result
	 * @param sqlFormat
	 * @param ...
	 * @return [bool] true on success
	 */
	bool query(SqliteResult &result, const char *sqlFormat, ...)
	{
		if (!this->isOpen()) {
			ELOG("libSqlite(%s): Database is not opened.\n", this->database.c_str());
			return false; /* failure */
		}

		/* prepare the sql query */
		char *sqlQuery;
		va_list sqlArgs;
		va_start(sqlArgs, sqlFormat);
		sqlQuery = sqlite3_vmprintf(sqlFormat, sqlArgs); /* must be freed */
		va_end(sqlArgs);

		int res = SQLITE_ERROR;
		for (int repeat = 0; repeat < this->sqliteRetries; repeat++) {
			result.reset();
			res = sqlite3_prepare_v2(this->sqlite, sqlQuery, -1, &result.statement, NULL);

			if (res != SQLITE_BUSY && res != SQLITE_LOCKED) {
				break; /* SQLITE_OK or an error */
			}

			/* retry */
			this->sqliteSleep();
		}

		if (res != SQLITE_OK) {
			ELOG("libSqlite(%s): Failed query \"%s\" error %d: %s.\n", this->database.c_str(), sqlQuery, sqlite3_errcode(this->sqlite), sqlite3_errmsg(this->sqlite));

			result.reset();
			sqlite3_free(sqlQuery);
			return false;
		}

		sqlite3_free(sqlQuery);
		return true; /* success */
	}

	/**
	 * Return
	 *  - true when a record is ready,
	 *  - false on error or when there is no more data to fetch
	 */
	bool fetchRow(SqliteResult &result)
	{
		int res = SQLITE_ERROR;

		if (!result.statement) {
			return res;
		}

		for (int repeat = 0; repeat < this->sqliteRetries; repeat++) {
			res = sqlite3_step(result.statement);

			switch (res) {
				case SQLITE_BUSY: /* retry */
				case SQLITE_LOCKED: /* retry */
					break;
				case SQLITE_ROW:
					return true; /* succes */
				case SQLITE_DONE:
					return false; /* success - end */
				default:
					ELOG("libSqlite(%s): Database fetch error %d: %s.\n", this->database.c_str(), sqlite3_errcode(this->sqlite), sqlite3_errmsg(this->sqlite));
					return false; /* error */
			}
			/* retry */
			this->sqliteSleep();
		}

		/* timeout */
		ELOG("libSqlite(%s): Database fetch timeout %d: %s.\n", this->database.c_str(), sqlite3_errcode(this->sqlite), sqlite3_errmsg(this->sqlite));

		return false;
	}

	bool dropTable(const std::string table)
	{
		if (!this->isOpen()) {
			ELOG("libSqlite(%s): Database is not opened.\n", this->database.c_str());
			return false; /* failure */
		}

		DLOG("libSqlite(%s): Dropping table \"%s\".\n", this->database.c_str(), table.c_str());

		return this->exec("drop table '%s'", table.c_str());
	}

	bool createTable(const std::string table, const std::string fields)
	{
		if (!this->isOpen()) {
			ELOG("libSqlite(%s): Database is not opened.\n", this->database.c_str());
			return false; /* failure */
		}

		std::string lockFile = std::string("/tmp/") + basename(database.c_str()) + ".lock";

		/* exclusive open, (blocking) */
		SqliteLock guard(lockFile);
		if (!guard.lock()) {
			return false;
		}

		/* look for the table while locked (probably again) */
		if (this->hasTable(table)) {
			DLOG("libSqlite(%s): Table \"%s\" already exists.\n", this->database.c_str(), table.c_str());
			return false;
		}

		/* create the table */
		bool res = this->exec("create table %s (%s);", table.c_str(), fields.c_str());
		if (res) {
			DLOG("libSqlite(%s): Table \"%s\" created.\n", this->database.c_str(), table.c_str());
		}
		else {

			ELOG("libSqlite(%s): Failed to create table \"%s\" .\n", this->database.c_str(), table.c_str());
		}
		return res;
	}

	/**
	 * Deletes the database AND the database journal
	 *
	 * @param dbFilename [string] - filename
	 */
	void dropDatabase(const std::string dbFilename)
	{
		this->close();

		std::string databaseJournal(dbFilename);
		databaseJournal += "-journal";

		unlink(dbFilename.c_str());
		unlink(databaseJournal.c_str());
	}

	int getChanges()
	{
		if (!this->isOpen()) {

			return 0;
		}

		return sqlite3_changes(this->sqlite);
	}

	bool hasTables(std::set<std::string> tables)
	{
		if (!this->isOpen()) {
			ELOG("libSqlite(%s): Database is not opened.\n", this->database.c_str());
			return false; /* failure */
		}

		SqliteResult result;
		if (!this->query(result, "SELECT name FROM sqlite_master WHERE type='table'")) {
			ELOG("libSqlite(%s): Failed to get tables from sqlite_master.\n", this->database.c_str());
			return false; /* error */
		}

		while (this->fetchRow(result)) {
			std::string table = (char *) sqlite3_column_text(result.statement, 0);
			tables.erase(table);
		}

		if (tables.size() > 0) {

			BOOST_FOREACH(std::string table, tables)
			{
				DLOG("libSqlite(%s): Table \"%s\" not found.\n", this->database.c_str(), table.c_str());
			}
			return false;
		}

		return true;
	}

	bool hasTable(const std::string table)
	{
		if (!this->isOpen()) {
			ELOG("libSqlite(%s): Database is not opened.\n", this->database.c_str());
			return false; /* failure */
		}

		SqliteResult result;
		if (!this->query(result, "SELECT name FROM sqlite_master WHERE type='table' AND tbl_name = '%s'", table.c_str())) {
			ELOG("libSqlite(%s): Failed to find table \"%s\" in sqlite_master.\n", this->database.c_str(), table.c_str());
			return false; /* error */
		}
		if (this->fetchRow(result)) {
			return true;
		}

		DLOG("libSqlite(%s): Table \"%s\" not found.\n", this->database.c_str(), table.c_str());

		return false;
	}

	bool hasTableColumn(const std::string table, const std::string column)
	{
		if (!this->isOpen()) {
			ELOG("libSqlite(%s): Database is not opened.\n", this->database.c_str());
			return false; /* failure */
		}

		SqliteResult result;
		this->query(result, "PRAGMA table_info(cdr)");
		while (this->fetchRow(result)) {
			if (column == std::string((char *) sqlite3_column_text(result.statement, 1))) {
				DLOG("libSqlite(%s): Table \"%s\" has column \"%s\".\n", this->database.c_str(), table.c_str(), column.c_str());
				return true;
			}
		}

		DLOG("libSqlite(%s): Table \"%s\" doesn't have column \"%s\".\n", this->database.c_str(), table.c_str(), column.c_str());
		return false;
	}
};

#endif /* LIBSQLITE_H */
