#include <fnmatch.h>

/**
 * Find string in a file
 */
static int helper_func_infile_read(struct ast_channel *chan, const char *cmd, char *data, char *buf, size_t len)
{
	char tmpBuf[BUFSIZ];
	char tmpFilename[BUFSIZ];

	AST_DECLARE_APP_ARGS(args,
		AST_APP_ARG(needle);
		AST_APP_ARG(filename);
		);

	if (ast_strlen_zero(data)) {
		ast_log(AST_LOG_WARNING, "Syntax: KERIO_IN_FILE(string,filename) - missing arguments!\n");
		return -1;
	}

	AST_STANDARD_APP_ARGS(args, data);

	if (args.argc != 2) {
		ast_log(AST_LOG_WARNING, "Syntax: KERIO_IN_FILE(string,filename) - missing arguments!\n");
		return -1;
	}

	if (ast_strlen_zero(args.filename)) {
		ast_log(AST_LOG_WARNING, "Syntax: KERIO_IN_FILE(string,filename) - missing arguments!\n");
		return -1;
	}

	if (ast_strlen_zero(args.needle)) {
		ast_verb(3, "Empty string.\n");
		snprintf(buf, len, "%d", 0);
		return 0;
	}


	/* prepare filename */
	if ('/' == *(args.filename)) {
		ast_copy_string(tmpFilename, args.filename, sizeof(tmpFilename));
	}
	else {
		snprintf(tmpFilename, sizeof(tmpFilename), "%s/%s", ast_config_AST_CONFIG_DIR, args.filename);
	}

	/* open the file */
	FILE *fp = fopen(tmpFilename, "r");
	if (NULL == fp) {
		ast_log(AST_LOG_ERROR, "Failed to open file %s.\n", tmpFilename);
		return -1;
	}

	while (fgets(tmpBuf, sizeof(tmpBuf), fp) != NULL) {
		/* remove line endings */
		if (tmpBuf[strlen(tmpBuf) - 1] == '\n') {
			tmpBuf[strlen(tmpBuf) - 1] = '\0';
		}

		/* compare */
		if (0 == strcmp(tmpBuf, args.needle)) {
			/* found */
			ast_verb(3, "Searching for '%s' in %s... [FOUND]\n", args.needle, tmpFilename);
			snprintf(buf, len, "%d", 1);
			fclose(fp);
			return 0;
		}
	}

	/* not found */
	ast_verb(3, "Searching for '%s' in %s... [NOT FOUND]\n", args.needle, tmpFilename);
	snprintf(buf, len, "%d", 0);
	fclose(fp);
	return 0;
}

static struct ast_custom_function helper_func_infile = {
	.name = "KERIO_IN_FILE",
	.desc = "Find string in a file",
	.syntax = "KERIO_IN_FILE(string,filename)",
	.read = helper_func_infile_read,
};

static int helper_func_peercount_read(struct ast_channel *chan, const char *cmd, char *data, char *buf, size_t len)
{
	AST_DECLARE_APP_ARGS(args,
		AST_APP_ARG(peer);
		);

	if (ast_strlen_zero(data)) {
		ast_log(AST_LOG_WARNING, "Syntax: KERIO_PEER_COUNT(i-peer) - missing arguments!\n");
		return -1;
	}

	AST_STANDARD_APP_ARGS(args, data);

	if (args.argc != 1) {
		ast_log(AST_LOG_WARNING, "Syntax: KERIO_PEER_COUNT(i-peer) - missing arguments!\n");
		return -1;
	}

	if (ast_strlen_zero(args.peer)) {
		ast_log(AST_LOG_WARNING, "Syntax: KERIO_PEER_COUNT(i-peer) - missing arguments!\n");
		return -1;
	}

	struct ast_channel *c = NULL;
	struct ast_channel_iterator *iter = NULL;

	ast_debug(3, "Iterating all channels... counting peer %s\n", args.peer);
	if (!(iter = ast_channel_iterator_all_new())) {
		ast_log(AST_LOG_ERROR, "Failed to get iterator for all channels.\n");
		return -1;
	}

	/* Count peers
	 *
	 * SIP channels return [category name] which equals to i-<interface name>-<something>
	 * i-bender-outbound
	 * i-bender-outbound-2
	 * i-bender-inbound-register1
	 * i-bender-inbound-register2
	 * i-bender-inbound
	 * i-bender-inbound-2
	 *
	 * Dahdi channels return context which equals to -i<interface name>:
	 * i-PRI
	 */

	char peerMatchSip[1024];
	snprintf(peerMatchSip, sizeof(peerMatchSip), "%s-+(outbound|inbound|inbound-register)*([-0-9])", args.peer);

	int count = 0;
	while ((c = ast_channel_iterator_next(iter))) {
		char optionData[1024];
		int optionDataLen = 1024 - 1;
		memset(optionData, 0, 1024);

		if (0 != ast_channel_queryoption(c, AST_OPTION_KERIO_CONTEXT, &optionData, &optionDataLen, 0)) {
			ast_debug(3, "Reading AST_OPTION_KERIO_CONTEXT failed for %s.\n", ast_channel_name(c));
			ast_channel_unref(c);
			continue;
		}

		ast_debug(3, "Name: %s context: %s\n", ast_channel_name(c), optionData);

		if (fnmatch(peerMatchSip, optionData, FNM_EXTMATCH) != FNM_NOMATCH) {
			/* SIP */
			ast_debug(3, "SIP peer %s matched\n", ast_channel_name(c));
			count++;
		}
		else if (0 == strcmp(optionData, args.peer)) {
			/* Dahdi */
			ast_debug(3, "Dahdi peer %s matched\n", ast_channel_name(c));
			count++;
		}
		ast_channel_unref(c);
	}

	ast_channel_iterator_destroy(iter);

	ast_debug(3, "Iterating all channels... counting peer %s, count=%d\n", args.peer, count);
	snprintf(buf, len, "%d", count);
	return 0;
}

static struct ast_custom_function helper_func_peercount = {
	.name = "KERIO_PEER_COUNT",
	.desc = "Count calls over a peer",
	.syntax = "KERIO_PEER_COUNT(i-peer)",
	.read = helper_func_peercount_read,
};

static int load_module_helper(void)
{
	ast_custom_function_register(&helper_func_infile);
	ast_custom_function_register(&helper_func_peercount);

	return AST_MODULE_LOAD_SUCCESS;
}

static int unload_module_helper(void)
{
	ast_custom_function_unregister(&helper_func_infile);
	ast_custom_function_unregister(&helper_func_peercount);

	return 0;
}

static int reload_module_helper(void)
{
	return 0;
}
