#ifndef ACTIVE_CALLS_H
#define	ACTIVE_CALLS_H

#include "Common.h"

class ActiveCalls {
private:
	int amiSchedulerId;
	struct timeval amiSchedulerDue;
	struct ast_sched_context *amiScheduler;

	/**
	 *  Asterisk Channel - ast_channel
	 */
	class AsteriskChannel {
	public:
		std::string uniqueId;
		std::string name;
		std::string codecRead;
		std::string codecWrite;
		std::string ktsMyPhoneTo;
		std::string ktsCallId;
		std::string ktsFromNum;
		std::string ktsFromName;
	};
	typedef std::map<std::string, AsteriskChannel> AsteriskChannels; /* key is uniqueId */
	typedef AsteriskChannels::iterator AsteriskChannelsIt;

	class Channel {
	public:
		std::string uniqueId;
		std::string channelName;

		int isLocal() const
		{
			if (channelName.empty() || !boost::starts_with(channelName, "Local/")) {
				return 0;
			}

			if (*channelName.rbegin() == '1') {
				return 1; /* first local channel */
			}

			return 2; /* second local channel */
		}

		bool isSip() const
		{
			return !channelName.empty() && boost::starts_with(channelName, "SIP/");
		}

		bool isHw() const
		{
			return !channelName.empty() && boost::starts_with(channelName, "DAHDI/");
		}

		std::string toStr() const
		{
			return uniqueId + "(" + channelName + ")";
		}

		bool operator==(const Channel &b) const
		{
			return uniqueId == b.uniqueId && channelName == b.channelName;
		}
	};
	typedef std::vector<Channel> ChannelList;
	typedef ChannelList::iterator ChannelListIt;

	class ActiveCall {
	public:
		std::string callId; /* Operator's unique call id */

		ChannelList channels;

		std::string fromName;
		std::string fromNumber;
		std::string fromNumberExternal;
		std::string fromInterface;
		int fromType;

		std::string toName;
		std::string toNumber;
		std::string toNumberExternal;
		std::string toInterface;
		int toType;

		std::string transferredByNumber;
		std::string parkedByNumber;
		int parkedByType;

		KTS_CallStatus callStatus;
		struct timeval timeCreated;
		struct timeval timeAnswered;

		std::string ktsCallId;

		int lastSeen;

		struct {
			int durationCreated;
			int durationAnswered;
			std::string fromCodecRead;
			std::string fromCodecWrite;
			std::string toCodecRead;
			std::string toCodecWrite;
		} computed;

		ActiveCall(const struct ast_cel_event_record *record) : fromType(0), toType(0), parkedByType(0), lastSeen(0)
		{
			callId = record->linked_id;

			Channel c;
			c.uniqueId = record->unique_id;
			c.channelName = record->channel_name;
			channels.push_back(c);

			fromName = record->caller_id_name;
			if (isNumberValid(record->caller_id_num)) {
				fromNumber = record->caller_id_num;
			}

			if (isNumberValid(record->extension)) {
				toNumber = record->extension;
			}

			timeCreated = record->event_time;
			timeAnswered.tv_sec = 0;
			timeAnswered.tv_usec = 0;

			callStatus = KTS_CallStatusConnecting;

			computed.durationCreated = 0;
			computed.durationAnswered = 0;

			updateLastSeen();
		}

		bool isConnected() const
		{
			if (this->channels.size() < 2) {
				return false;
			}

			return(!this->channels.begin()->isLocal() && !this->channels.rbegin()->isLocal());
		}

		void appendRecord(const struct ast_cel_event_record *record)
		{
			Channel c;
			c.uniqueId = record->unique_id;
			c.channelName = record->channel_name;
			this->channels.push_back(c);

			if (strlen(record->caller_id_name) > 0 && 0 != strcmp(record->caller_id_name, "s")) {
				this->toName = record->caller_id_name;
			}
			if (isNumberValid(record->caller_id_num)) {
				this->toNumber = record->caller_id_num;
			}

			if (!isNumberValid(this->fromNumber) && isNumberValid(record->extension)) {
				this->fromNumber = record->extension;
			}
		}

		void prependRecord(const struct ast_cel_event_record *record)
		{
			Channel c;
			c.uniqueId = record->unique_id;
			c.channelName = record->channel_name;
			this->channels.insert(this->channels.begin(), c);

			if (strlen(record->caller_id_name) > 0 && 0 != strcmp(record->caller_id_name, "s")) {
				this->fromName = record->caller_id_name;
			}
			if (isNumberValid(record->caller_id_num)) {
				this->fromNumber = record->caller_id_num;
			}

			if (!isNumberValid(this->toNumber) && isNumberValid(record->extension)) {
				this->toNumber = record->extension;
			}
		}

		void prependLast(ActiveCall &c)
		{
			fromName = c.toName;
			fromNumber = c.toNumber;
			fromNumberExternal = c.toNumberExternal;
			fromType = c.toType;
			fromInterface = c.toInterface;

			channels.insert(channels.begin(), c.channels.back());
		}

		void appendCall(ActiveCall &c)
		{
			toName = c.toName;
			toNumber = c.toNumber;
			toNumberExternal = c.toNumberExternal;
			toType = c.toType;
			toInterface = c.toInterface;

			channels.insert(channels.end(), c.channels.begin(), c.channels.end());
		}

		void swapLast(ActiveCall &c)
		{
			toName.swap(c.toName);
			toNumber.swap(c.toNumber);
			toNumberExternal.swap(c.toNumberExternal);
			std::swap(toType, c.toType);
			toInterface.swap(c.toInterface);

			if (!channels.empty() && !c.channels.empty()) {
				Channel last1 = *channels.rbegin();
				Channel last2 = *c.channels.rbegin();
				channels.pop_back();
				c.channels.pop_back();
				channels.push_back(last2);
				c.channels.push_back(last1);
			}
		}

		void reverse()
		{
			fromName.swap(toName);
			fromNumber.swap(toNumber);
			fromNumberExternal.swap(toNumberExternal);
			std::swap(fromType, toType);
			fromInterface.swap(toInterface);

			std::reverse(channels.begin(), channels.end());

			/* Local channels should remain in the correct order */
			for (ChannelListIt it = channels.begin(); it != channels.end(); ++it) {
				if (it->isLocal() == 2 && (it + 1) != channels.end() && (it + 1)->isLocal() == 1) {
					std::iter_swap(it, it + 1);
					it++;
				}
			}
		}

		void compute()
		{
			/* durations */
			const struct timeval now = ast_tvnow();
			this->computed.durationAnswered = ast_tvzero(this->timeAnswered) ? 0 : ast_tvdiff_ms(now, this->timeAnswered);
			this->computed.durationCreated = ast_tvzero(this->timeCreated) ? 0 : ast_tvdiff_ms(now, this->timeCreated);
		}

		void compute(AsteriskChannels &asteriskChannels)
		{
			this->compute();

			int channelCounter = 0;

			this->computed.fromCodecRead = "";
			this->computed.fromCodecWrite = "";
			this->computed.toCodecRead = "";
			this->computed.toCodecWrite = "";

			BOOST_FOREACH(const Channel &channel, this->channels)
			{
				AsteriskChannelsIt asteriskChannelsIt = asteriskChannels.find(channel.uniqueId);
				if (asteriskChannelsIt == asteriskChannels.end()) {
					continue;
				}

				/* codecs */
				if (channel.isSip()) {
					if (channelCounter == 0) {
						this->computed.fromCodecRead = asteriskChannelsIt->second.codecRead;
						this->computed.fromCodecWrite = asteriskChannelsIt->second.codecWrite;
					}
					else {
						this->computed.toCodecRead = asteriskChannelsIt->second.codecRead;
						this->computed.toCodecWrite = asteriskChannelsIt->second.codecWrite;
					}
				}

				/* myphone dialing */
				if (!asteriskChannelsIt->second.ktsMyPhoneTo.empty() && !isNumberValid(this->toNumber)) {
					this->callStatus = KTS_CallStatusPickUp;
					this->toNumber = asteriskChannelsIt->second.ktsMyPhoneTo;
				}

				/* faxes from myphone */
				if (isNumberValid(asteriskChannelsIt->second.ktsFromNum)) {
					this->fromNumber = asteriskChannelsIt->second.ktsFromNum;
					this->fromName = asteriskChannelsIt->second.ktsFromName;
				}

				if (!asteriskChannelsIt->second.ktsCallId.empty() && this->ktsCallId.empty()) {
					this->ktsCallId = asteriskChannelsIt->second.ktsCallId;
				}

				channelCounter++;
			}
		}

		void updateLastSeen()
		{
			struct timespec tv;
			clock_gettime(CLOCK_MONOTONIC, &tv);
			this->lastSeen = tv.tv_sec;
		}

		int getLastSeen() const
		{
			if (this->lastSeen <= 0) {
				return 0;
			}
			struct timespec tv;
			clock_gettime(CLOCK_MONOTONIC, &tv);
			return tv.tv_sec - this->lastSeen;
		}

		void deleteChannel(const std::string &uniqueId)
		{
			for (ChannelListIt it = this->channels.begin(); it != this->channels.end();) {
				if (it->uniqueId == uniqueId) {
					it = this->channels.erase(it);
				}
				else {
					++it;
				}
			}
		}

		bool channelExists(const std::string &uniqueId)
		{
			for (ChannelListIt it = this->channels.begin(); it != this->channels.end();) {
				if (it->uniqueId == uniqueId) {
					return true;
				}
			}

			return false;
		}

		std::string getChannels() const
		{
			std::stringstream ss;
			bool first = true;

			BOOST_FOREACH(const Channel &chan, this->channels)
			{
				if (!first) {
					ss << "&";
				}
				ss << chan.toStr();
				first = false;
			}

			return ss.str();
		}
	};
	typedef std::vector<ActiveCall> ActiveCallList;
	typedef ActiveCallList::iterator ActiveCallListIt;
	typedef ActiveCallList::const_iterator ActiveCallListConstIt;
	ActiveCallList activeCallList;
	Mutex activeCallListMutex;


	typedef std::map<std::string, ActiveCallsExtension> ActiveCallsExtensionList; /* key is telNum */
	ActiveCallsExtensionList activeCallsExtensionList;

	ast_json *getCall(ActiveCallListConstIt &it);
	void debugPrintCalls();
	void deleteCall(ActiveCallListIt it);
	bool findChannel(ActiveCallListIt &it, const std::string &uniqueId, ChannelListIt *itChan = NULL, bool *first = NULL, bool *last = NULL);
	std::string generateCallId(const std::string &callId);
	void cleanup();
	std::string getCodecDescription(const struct ast_format *format);
	AsteriskChannel getAsteriskChannel(struct ast_channel *c);
	AsteriskChannels getAsteriskChannels();
	void compute();

public:

	ActiveCalls();
	~ActiveCalls();
	void reload();
	void print(const int fd);
	void cel(struct ast_event *event);
	void startAmiScheduler();
	void amiSendCalls();

	KTS_ExtensionType getExtensionTypeConf(const std::string &telNum);
	KTS_ExtensionType getExtensionType(const std::string &value);

	static std::string getChannelVariable(const std::string paramName, struct ast_channel *chan, const std::string defaultValue = "");
	static void setChannelVariable(struct ast_channel *chan, const std::string paramName, const std::string paramValue);
	static std::string getCdrVariable(const std::string &name, const std::string &channelName, const std::string defaultValue = "");

	int hangupOutgoingCalls(std::vector<std::string> emergencyList, std::vector<std::string> routePrefixes);
	int getCallsDurationPrefix(int scope, std::vector<std::string> prefixes);
	int getCallsCountPrefix(int scope, std::vector<std::string> prefixes);
};

#endif	/* ACTIVE_CALLS_H */
